import inspect
from typing import Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss as Loss

from serotiny.models.base_model import BaseModel
from .priors import Prior, IsotropicGaussianPrior

Array = Union[torch.Tensor, np.ndarray, Sequence[float]]


def get_args(encoder):
    args = []
    if isinstance(encoder, nn.Sequential):
        for i in range(len(encoder)):
            args.append(inspect.getfullargspec(encoder[i].forward).args)
        args = [item for sublist in args for item in sublist]
    else:
        args = inspect.getfullargspec(encoder.forward).args
    return args


class BaseVAE(BaseModel):
    def __init__(
        self,
        encoder: nn.Module,
        decoder: nn.Module,
        latent_dim: Union[int, Sequence[int]],
        beta: float,
        x_label: str,
        id_label: Optional[str] = None,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        loss_mask_label: Optional[str] = None,
        reconstruction_loss: Loss = nn.MSELoss(reduction="none"),
        reconstruction_reduce: str = "sum",
        priors: Optional[Sequence[Prior]] = None,
        cache_outputs: Sequence = ("test",),
        **kwargs,
    ):
        """Instantiate a basic VAE model.

        Parameters
        ----------
        encoder: nn.Module
            Encoder network
        decoder: nn.Module
            Decoder network
        optimizer: torch.optim.Optimizer
            Optimizer to use
        beta: float
            Beta parameter - the weight of the KLD term in the loss function
        x_label: str
            String label used to retrieve X from the batches
        reconstruction_loss: Loss
            Loss to be used for reconstruction. Can be a PyTorch loss or a class
            that respects the same interface,
            i.e. subclasses torch.nn.modules._Loss
        priors: Optional[Sequence[AbstractPrior]]
            List of prior specifications to use for latent space
        """
        super().__init__()

        self.reconstruction_reduce = reconstruction_reduce
        self.reconstruction_loss = reconstruction_loss
        self.encoder = encoder
        self.decoder = decoder

        self.beta = beta
        self.latent_dim = latent_dim

        self.encoder_args = inspect.getfullargspec(self.encoder.forward).args
        self.decoder_args = inspect.getfullargspec(self.decoder.forward).args

        if priors is None:
            priors = [IsotropicGaussianPrior()]

        self.priors = nn.ModuleList(priors)

    def calculate_elbo(self, x, x_hat, z, mask=None):
        rcl_per_input_dimension = self.reconstruction_loss(x_hat, x)

        if mask is not None:
            rcl_per_input_dimension = rcl_per_input_dimension * mask
            normalizer = mask.view(mask.shape[0], -1).sum(dim=1)
        else:
            normalizer = np.prod(x.shape[1:])

        rcl = (
            rcl_per_input_dimension
            # flatten
            .view(rcl_per_input_dimension.shape[0], -1)
            # and sum across each batch element's dimensions
            .sum(dim=1)
        )

        if self.reconstruction_reduce == "mean":
            rcl = rcl / normalizer

        rcl = rcl.mean()

        kld_per_dimension = torch.cat(
            [prior(z, reduction="none", mode="kl") for prior in self.priors], dim=1
        )

        kld = kld_per_dimension.sum(dim=1).mean()

        return (
            rcl + self.beta * kld,
            rcl,
            kld,
            rcl_per_input_dimension,
            kld_per_dimension,
        )

    def parse_batch(self, batch):
        if self.hparams.loss_mask_label is not None:
            mask = batch[self.hparams.loss_mask_label].float()
        else:
            mask = None
        return batch[self.hparams.x_label].float(), dict(mask=mask)

    def sample_z(self, z):
        return torch.cat([prior(z, mode="sample") for prior in self.priors], dim=1)

    def encode(self, x, **kwargs):
        return self.encoder(
            x, **{k: v for k, v in kwargs.items() if k in self.encoder_args}
        )

    def decode(self, z, **kwargs):
        return self.decoder(
            z, **{k: v for k, v in kwargs.items() if k in self.decoder_args}
        )

    def forward(self, x, decode=False, compute_loss=False, **kwargs):
        z_params = self.encode(x, **kwargs)
        if not decode:
            return z_params

        z = self.sample_z(z_params)
        x_hat = self.decode(z, **kwargs)

        if not compute_loss:
            return x_hat, z, z_params

        (
            loss,
            reconstruction_loss,
            kld_loss,
            rcl_per_input_dimension,
            kld_per_latent_dimension,
        ) = self.calculate_elbo(x, x_hat, z_params, mask=kwargs.get("mask", None))

        return (
            x_hat,
            z,
            z_params,
            loss,
            reconstruction_loss,
            kld_loss,
            rcl_per_input_dimension,
            kld_per_latent_dimension,
        )

    def log_metrics(self, stage, reconstruction_loss, kld_loss, loss, logger):
        on_step = stage == "train"

        self.log(
            f"{stage} reconstruction loss",
            reconstruction_loss,
            logger=logger,
            on_step=on_step,
            on_epoch=True,
        )
        self.log(
            f"{stage} kld loss", kld_loss, logger=logger, on_step=on_step, on_epoch=True
        )
        self.log(f"{stage}_loss", loss, logger=logger, on_step=on_step, on_epoch=True)

    def _step(self, stage, batch, batch_idx, logger):

        x = self.parse_batch(batch)

        if isinstance(x, tuple):
            x, forward_kwargs = x
        else:
            forward_kwargs = dict()

        (
            x_hat,
            z,
            z_params,
            loss,
            reconstruction_loss,
            kld_loss,
            rcl_per_input_dimension,
            kld_per_latent_dimension,
        ) = self.forward(x, decode=True, compute_loss=True, **forward_kwargs)

        self.log_metrics(stage, reconstruction_loss, kld_loss, loss, logger)

        results = {
            "loss": loss,
            f"{stage}_loss": loss.detach().cpu(),  # for epoch end logging purposes
            "reconstruction_loss": reconstruction_loss.detach().cpu(),
            "kld_loss": kld_loss.detach().cpu(),
            "batch_idx": batch_idx,
            "z": z.detach().cpu(),
            "z_params": z_params.detach().cpu(),
            "kld_per_latent_dimension": kld_per_latent_dimension.detach().float().cpu(),
            "rcl_per_input_dimension": rcl_per_input_dimension.detach().float().cpu(),
        }

        if stage == "test":
            results.update(
                {
                    "x_hat": x_hat.detach().cpu(),
                    "x": x.detach().cpu(),
                }
            )

        if self.hparams.id_label is not None:
            if self.hparams.id_label in batch:
                ids = batch[self.hparams.id_label].detach().cpu()
                results.update({self.hparams.id_label: ids, "id": ids})

        return results



import re
from itertools import chain
from upath import UPath as Path
from typing import Dict, Optional, Sequence, Union

import numpy as np
import pandas as pd
import pytorch_lightning as pl
from pytorch_lightning.utilities.exceptions import MisconfigurationException
from torch.utils.data import DataLoader

from serotiny.io.dataframe import DataframeDataset, read_dataframe
from serotiny.io.dataframe.loaders.abstract_loader import Loader
from .utils import FastDataLoader


class ManifestDatamodule(pl.LightningDataModule):
    """A pytorch lightning datamodule based on manifest files. It can either
    use a single manifest file, which contains a column based on which a train-
    val-test split can be made; or it can use three manifest files, one for
    each fold (train, val, test).

    Additionally, if it is only going to be used for prediction/testing, a flag
    `just_inference` can be set to True so the splits are ignored and the whole
    dataset is used.

    The `predict_datamodule` is simply constructed from the whole dataset,
    regardless of the value of `just_inference`.
    """

    def __init__(
        self,
        path: Union[Path, str],
        loaders: Union[Dict, Loader],
        split_column: Optional[Union[Path, str]] = None,
        columns: Optional[Sequence[str]] = None,
        just_inference: bool = False,
        dataloader_type: str = "fast",
        **dataloader_kwargs,
    ):
        """
        Parameters
        ----------
        path: Union[Path, str]
            Path to a manifest file

        loaders: Union[Dict, Loader]
            Loader specifications for each given split.

        split_column: Optional[str] = None
            Name of a column in the dataset which can be used to create train, val, test
            splits.

        columns: Optional[Sequence[str]] = None
            List of columns to load from the dataset, in case it's a parquet file.
            If None, load everything.

        just_inference: bool = False
            Whether this datamodule will be used for just inference
            (testing/prediction).
            If so, the splits are ignored and the whole dataset is used.

        dataloader_kwargs:
            Additional keyword arguments are passed to the
            torch.utils.data.DataLoader class when instantiating it (aside from
            `shuffle` which is only used for the train dataloader).
            Among these args are `num_workers`, `batch_size`, `shuffle`, etc.
            See the PyTorch docs for more info on these args:
            https://pytorch.org/docs/stable/data.html#torch.utils.data.DataLoader
        """

        super().__init__()
        self.path = path
        self._dataloader_cls = (
            FastDataLoader if dataloader_type == "fast" else DataLoader
        )

        # if only one loader is specified, the same loaders are
        # used for all  folds
        loaders = _parse_loaders(loaders)

        path = Path(path)

        # if path is a directory, we assume it is a directory containing multiple
        # manifests - one per split. otherwise, we assume it is the path to a single
        # manifest file, which is expected to have a
        if path.is_dir():
            self.datasets = _make_multiple_manifest_splits(
                path, loaders, columns, just_inference
            )
        else:
            if split_column is None and not just_inference:
                raise MisconfigurationException(
                    "When using a single manifest file, it must have a "
                    "split column, to use for train-val-test splitting."
                )

            self.datasets = _make_single_manifest_splits(
                path, loaders, split_column, columns, just_inference
            )

        self.just_inference = just_inference
        self.dataloader_kwargs = dataloader_kwargs
        self.dataloaders = dict()

    def make_dataloader(self, split):
        kwargs = dict(**self.dataloader_kwargs)
        kwargs["shuffle"] = kwargs.get("shuffle", True) and split == "train"

        return self._dataloader_cls(dataset=self.datasets[split], **kwargs)

    def _get_dataloader(self, split):
        if split not in self.dataloaders:
            self.dataloaders[split] = self.make_dataloader(split)

        return self.dataloaders[split]

    def train_dataloader(self):
        if self.just_inference:
            raise TypeError(
                "This datamodule was configured with `just_inference=True`, "
                "so it doesn't have a train_dataloader and can't be "
                "used for training."
            )
        return self._get_dataloader("train")

    def val_dataloader(self):
        if self.just_inference:
            raise TypeError(
                "This datamodule was configured with `just_inference=True`, "
                "so it doesn't have a train_dataloader and can't be "
                "used for training."
            )

        return self._get_dataloader("val")

    def test_dataloader(self):
        split = "predict" if self.just_inference else "test"
        return self._get_dataloader(split)

    def predict_dataloader(self):
        return self._get_dataloader("predict")


def _get_canonical_split_name(split):
    for canon in ["train", "val", "test", "predict"]:
        if split.startswith(canon) or canon.startswith(split):
            return canon
    raise ValueError


def _make_single_manifest_splits(
    manifest_path, loaders, split_column, columns=None, just_inference=False
):
    dataframe = read_dataframe(manifest_path, columns)
    if not just_inference:
        assert dataframe.dtypes[split_column] == np.dtype("O")

    split_names = dataframe[split_column].unique().tolist()
    if not just_inference:
        assert set(split_names).issubset(
            {"train", "training", "valid", "val", "validation", "test", "testing"}
        )

    if split_column != "split":
        dataframe["split"] = dataframe[split_column].apply(_get_canonical_split_name)

    if not just_inference:
        datasets = {}
        for split in ["train", "val", "test"]:
            datasets[split] = DataframeDataset(
                dataframe.loc[dataframe["split"].str.startswith(split)].copy(),
                loaders[split],
            )

    datasets["predict"] = DataframeDataset(dataframe.copy(), loaders["predict"])
    return datasets


def _make_multiple_manifest_splits(
    split_path, loaders, columns=None, just_inference=False
):
    split_path = Path(split_path)
    datasets = {}
    predict_df = []

    for fpath in chain(split_path.glob("*.csv"), split_path.glob("*.parquet")):
        split = re.findall(r"(.*)\.(?:csv|parquet)", fpath.name)[0]
        split = _get_canonical_split_name(split)
        dataframe = read_dataframe(fpath, required_columns=columns)
        dataframe["split"] = split
        dataset = DataframeDataset(dataframe, loaders=loaders[split])
        if not just_inference:
            datasets[split] = dataset
        predict_df.append(dataframe.copy())

    predict_df = pd.concat(predict_df)
    datasets["predict"] = DataframeDataset(predict_df, loaders=loaders["predict"])

    return datasets


def _dict_depth(d):
    return max(_dict_depth(v) if isinstance(v, dict) else 0 for v in d.values()) + 1


def _parse_loaders(loaders):
    depth = _dict_depth(loaders)
    if depth == 1:
        loaders = {split: loaders for split in ["train", "val", "test", "predict"]}
    elif depth != 2:
        raise ValueError(f"Loaders dict should have depth 1 or 2. Got {depth}")

    for k, v in loaders.items():
        loaders[_get_canonical_split_name(k)] = v

    for k in loaders:
        if isinstance(loaders[k], str):
            assert loaders[k] in loaders
            loaders[k] = loaders[loaders[k]]

    for split in ["train", "val", "test"]:
        if split not in loaders:
            raise ValueError(f"'{split}' missing from loaders dict.")

    if "predict" not in loaders:
        loaders["predict"] = loaders["test"]

    return loaders



"""Taken from https://github.com/PyTorchLightning/pytorch-
lightning/issues/1506#issue-600753413."""

import torch.utils.data


class _RepeatSampler(object):
    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            yield from iter(self.sampler)


class FastDataLoader(torch.utils.data.dataloader.DataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, "batch_sampler", _RepeatSampler(self.batch_sampler))
        self.iterator = super().__iter__()

    def __len__(self):
        return len(self.batch_sampler.sampler)

    def __iter__(self):
        for _ in range(len(self)):
            yield next(self.iterator)



def info(path):
    from serotiny.io.image import image_loader

    img, channels = image_loader(path, return_channels=True, return_as_torch=False)
    print("Image shape: ", img.shape)
    print("Channel names: ", list(channels.keys()))



from typing import Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn

from torch.nn.modules.loss import _Loss as Loss

from serotiny.networks.mlp import MLP

from .base_vae import BaseVAE
from .priors import Prior

Array = Union[torch.Tensor, np.ndarray, Sequence[float]]


class TabularVAE(BaseVAE):
    def __init__(
        self,
        x_dim: int,
        latent_dim: int,
        hidden_layers: Sequence[int],
        x_label: str,
        id_label: Optional[str] = None,
        loss_mask_label: Optional[str] = None,
        optimizer=torch.optim.Adam,
        beta: float = 1.0,
        priors: Optional[Sequence[Prior]] = None,
        reconstruction_loss: Loss = nn.MSELoss(reduction="none"),
        reconstruction_reduce: str = "sum",
        cache_outputs: Sequence = ("test",),
    ):
        encoder = MLP(
            x_dim,
            2 * latent_dim,
            hidden_layers=hidden_layers,
        )

        decoder = MLP(
            latent_dim,
            x_dim,
            hidden_layers=hidden_layers,
        )

        super().__init__(
            encoder=encoder,
            decoder=decoder,
            latent_dim=latent_dim,
            optimizer=optimizer,
            x_label=x_label,
            id_label=id_label,
            loss_mask_label=loss_mask_label,
            beta=beta,
            priors=priors,
            cache_outputs=cache_outputs,
            reconstruction_loss=reconstruction_loss,
            reconstruction_reduce=reconstruction_reduce,
        )



import logging
from typing import Optional, Sequence, Union

import numpy as np
import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss as Loss

from serotiny.networks import BasicCNN
from serotiny.networks.vae import ImageDecoder
from serotiny.networks.utils import weight_init

from .base_vae import BaseVAE
from .priors import Prior

Array = Union[torch.Tensor, np.ndarray, Sequence[float]]
logger = logging.getLogger("lightning")
logger.propagate = False


class ImageVAE(BaseVAE):
    def __init__(
        self,
        latent_dim: Union[int, Sequence[int]],
        in_channels: int,
        hidden_channels: Sequence[int],
        max_pool_layers: Sequence[int],
        input_dims: Sequence[int],
        x_label: str,
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        beta: float = 1.0,
        id_label: Optional[str] = None,
        non_linearity: Optional[nn.Module] = None,
        decoder_non_linearity: Optional[nn.Module] = None,
        loss_mask_label: Optional[str] = None,
        reconstruction_loss: Loss = nn.MSELoss(reduction="none"),
        reconstruction_reduce: str = "mean",
        skip_connections: bool = True,
        batch_norm: bool = True,
        mode: str = "3d",
        priors: Optional[Sequence[Prior]] = None,
        kernel_size: int = 3,
        cache_outputs: Sequence = ("test",),
        final_non_linearity: nn.Module = nn.Threshold(6, 6),
    ):

        encoder = BasicCNN(
            in_channels=in_channels,
            input_dims=input_dims,
            output_dim=latent_dim * 2,  # because it has to return mu and sigma
            hidden_channels=hidden_channels,
            max_pool_layers=max_pool_layers,
            mode=mode,
            kernel_size=kernel_size,
            non_linearity=non_linearity,
            skip_connections=skip_connections,
            batch_norm=batch_norm,
            final_non_linearity=final_non_linearity,
        )
        encoder.apply(weight_init)
        nn.utils.spectral_norm(encoder.output)

        dummy_out, intermediate_sizes = encoder.conv_forward(
            torch.zeros(1, in_channels, *input_dims), return_sizes=True
        )

        compressed_img_shape = dummy_out.shape[2:]

        intermediate_sizes = [input_dims] + intermediate_sizes[:-1]
        intermediate_sizes = intermediate_sizes[::-1]

        decoder = ImageDecoder(
            compressed_img_shape=compressed_img_shape,
            hidden_channels=list(reversed(hidden_channels)),
            intermediate_sizes=intermediate_sizes,
            latent_dim=latent_dim,
            output_dims=tuple(input_dims),
            output_channels=in_channels,
            mode=mode,
            non_linearity=non_linearity,
            skip_connections=skip_connections,
            batch_norm=batch_norm,
            kernel_size=kernel_size,
        )
        decoder.apply(weight_init)
        nn.utils.spectral_norm(decoder.linear_decompress)

        if decoder_non_linearity is not None:
            decoder = nn.Sequential(decoder, decoder_non_linearity)

        super().__init__(
            encoder=encoder,
            decoder=decoder,
            latent_dim=latent_dim,
            optimizer=optimizer,
            x_label=x_label,
            id_label=id_label,
            loss_mask_label=loss_mask_label,
            beta=beta,
            reconstruction_loss=reconstruction_loss,
            reconstruction_reduce=reconstruction_reduce,
            priors=priors,
            cache_outputs=cache_outputs,
        )



import inspect

import torch.optim as opt


def find_optimizer(optimizer_name):
    """Given optimizer name, get it from torch.optim."""
    available_optimizers = []
    for cls_name, cls in opt.__dict__.items():
        if inspect.isclass(cls):
            if issubclass(cls, opt.Optimizer):
                available_optimizers.append(cls_name)

    if optimizer_name in available_optimizers:
        optimizer_class = opt.__dict__[optimizer_name]
    else:
        raise KeyError(
            f"optimizer {optimizer_name} not available, "
            f"options are {available_optimizers}"
        )
    return optimizer_class


def find_lr_scheduler(scheduler_name):
    """Given scheduler name, get it from torch.optim.lr_scheduler."""
    available_schedulers = []
    for cls_name, cls in opt.lr_scheduler.__dict__.items():
        if inspect.isclass(cls):
            if "LR" in cls_name and cls_name[0] != "_":
                available_schedulers.append(cls_name)

    if scheduler_name in available_schedulers:
        scheduler_class = opt.lr_scheduler.__dict__[scheduler_name]
    else:
        raise Exception(
            f"scheduler {scheduler_name} not available, "
            f"options are {available_schedulers}"
        )
    return scheduler_class



import logging
import os
import tempfile
from pathlib import Path
from contextlib import contextmanager

import mlflow
from hydra._internal.utils import _locate
from mlflow.tracking import MlflowClient
from mlflow.utils.autologging_utils import autologging_integration
from omegaconf import OmegaConf

from .utils import flatten_config, save_model_predictions
from .ml_ops import instantiate

logger = logging.getLogger(__name__)


@autologging_integration("pytorch")
def patched_autolog(
    log_every_n_epoch=1,
    log_models=False,  # we handle this
    disable=False,
    exclusive=False,
    disable_for_unsupported_versions=False,
    silent=False,
    mode="fit",
):
    """Patched mlflow autolog, to enable on_test_epoch_end callbacks and to
    cover trainer.test calls as well."""

    import mlflow.pytorch._pytorch_autolog as _pytorch_autolog
    from mlflow.utils.autologging_utils import safe_patch
    from pytorch_lightning import Trainer
    from pytorch_lightning.utilities import rank_zero_only

    @rank_zero_only
    def _patched_on_test_epoch_end(
        original, self, trainer, model, *args
    ):  # pylint: disable=signature-differs,arguments-differ,unused-argument
        self._log_metrics(trainer, model, *args)

    assert mode in ["fit", "test", "predict"]

    safe_patch(
        "pytorch",
        _pytorch_autolog.__MLflowPLCallback,
        "on_test_epoch_end",
        _patched_on_test_epoch_end,
    )

    safe_patch("pytorch", mlflow.pytorch, "autolog", patched_autolog)

    safe_patch("pytorch", Trainer, mode, _pytorch_autolog.patched_fit, manage_run=True)

    safe_patch("pytorch", Trainer, "save_checkpoint", _patched_save_checkpoint)


def _is_empty(conf, key):
    return conf.get(key, None) is None


def _validate_mlflow_conf(conf):
    if _is_empty(conf, "tracking_uri"):
        raise ValueError("Must specify `tracking_uri`")

    if _is_empty(conf, "experiment_name") and _is_empty(conf, "run_id"):
        raise ValueError(
            "If you don't specify `experiment_name`, you must " "specify `run_id`."
        )

    if _is_empty(conf, "run_name") and _is_empty(conf, "run_id"):
        raise ValueError("You must specify at least `run_id` or `run_name`")


def _patched_save_checkpoint(original, self, filepath, save_weights_only):
    original(self, filepath, save_weights_only)

    latest_path = Path(filepath).with_name("latest.ckpt")
    os.link(filepath, latest_path)

    mlflow.log_artifact(local_path=latest_path, artifact_path="checkpoints")
    os.unlink(latest_path)


def _get_latest_checkpoint(tracking_uri, run_id, tmp_dir):
    client = MlflowClient(tracking_uri=tracking_uri)

    try:
        return client.download_artifacts(
            run_id=run_id, path="checkpoints/latest.ckpt", dst_path=tmp_dir
        )
    except:  # noqa
        return None


def _get_config(tracking_uri, run_id, tmp_dir, mode="test"):
    client = MlflowClient(tracking_uri=tracking_uri)

    try:
        return client.download_artifacts(
            run_id=run_id, path=f"configs/{mode}_config.yaml", dst_path=tmp_dir
        )
    except:  # noqa
        return None


def load_model_from_checkpoint(tracking_uri, run_id):
    mlflow.set_tracking_uri(tracking_uri)
    with tempfile.TemporaryDirectory() as tmp_dir:
        ckpt_path = _get_latest_checkpoint(tracking_uri, run_id, tmp_dir)
        if ckpt_path is None:
            logger.info("Given run_id doesn't have a checkpoint we can use.")
            return None

        config = _get_config(tracking_uri, run_id, tmp_dir, mode="train")
        config = OmegaConf.load(config)
        config = OmegaConf.to_container(config, resolve=True)

        model_conf = config["model"]
        model_class = model_conf.pop("_target_")
        model_conf = instantiate(model_conf)
        model_class = _locate(model_class)
        return model_class.load_from_checkpoint(ckpt_path, **model_conf)


def _get_patience(trainer):
    from pytorch_lightning.callbacks import EarlyStopping

    for callback in trainer.callbacks:
        if isinstance(callback, EarlyStopping):
            return callback.patience
    return None


def _mlflow_prep(mlflow_conf, trainer, mode):
    logger.info("Validating and processing MLFlow configuration")

    _validate_mlflow_conf(mlflow_conf)
    assert mode in ["fit", "test", "predict"]

    # if autolog arguments aren't given, or are None, set to empty dict
    autolog = (
        OmegaConf.to_object(mlflow_conf.autolog)
        if hasattr(mlflow_conf, "autolog")
        else None
    )
    autolog = autolog if autolog is not None else {}
    autolog["mode"] = mode
    if mode == "test":
        autolog["log_models"] = False
    patched_autolog(**autolog)

    mlflow.set_tracking_uri(mlflow_conf.tracking_uri)

    run_id = mlflow_conf.get("run_id", None)
    if run_id is None:
        run_name = mlflow_conf.get("run_name", None)
        assert run_name is not None
        # creates experiment if it doesn't exist, otherwise just gets it
        experiment = mlflow.set_experiment(experiment_name=mlflow_conf.experiment_name)

        runs = []
        for run_info in mlflow.list_run_infos(experiment_id=experiment.experiment_id):
            run_tags = mlflow.get_run(run_info.run_id).data.tags

            if "mlflow.runName" in run_tags:
                if run_tags["mlflow.runName"] == run_name:
                    runs.append(run_info.run_id)

        if len(runs) > 1:
            raise ValueError(
                "You provided `run_name`, but there are multiple "
                "runs in this experiment with that name. Please "
                "specify `run_id`."
            )
        elif len(runs) == 1:
            run_id = runs[0]
        else:  # redundant but leaving it here to be explicit
            run_id = None

    else:
        run = mlflow.get_run(run_id=run_id)
        experiment = mlflow.set_experiment(experiment_id=run.info.experiment_id)

    patience = _get_patience(trainer)

    # before this point, argv[0] will contain e.g. "/path/to/exectuable/serotiny train"
    # because we append "train" or "test" for presentation purposes in the CLI.
    # however, this causes problems when using tools that need to fork this program
    # e.g. torch's ddp training strategy, and which make use of argv[0] to rerun
    # the program. For that reason, from this point on we remove the suffix, and
    # re-insert it as the second element in argv

    try:
        import sys

        command, suffix = sys.argv[0].split(" ")
        sys.argv[0] = command
        sys.argv.insert(1, suffix)
    except:  # noqa
        # in hydra sweeps, this only needs to be done once and will cause an
        # error if we try to do it again
        pass

    return experiment, run_id, patience


def _log_conf(tmp_dir, full_conf, mode):
    if mode not in ("train", "test", "predict"):
        raise ValueError(f"Got unexpected mode: '{mode}")

    conf_path = Path(tmp_dir) / f"{mode}_config.yaml"
    with open(conf_path, "w") as f:
        OmegaConf.save(config=full_conf, f=f)
        mlflow.log_artifact(local_path=conf_path, artifact_path="configs")


def _log_predictions(preds_dir, tracking_uri, run_id):
    client = MlflowClient(tracking_uri=tracking_uri)
    client.log_artifacts(
        run_id=run_id, local_dir=preds_dir, artifact_path="predictions"
    )


def mlflow_fit(mlflow_conf, trainer, model, data, full_conf, test=False, predict=False):
    experiment, run_id, patience = _mlflow_prep(mlflow_conf, trainer, "fit")

    # if run_id has been specified, we're trying to resume
    if run_id is not None and not trainer.checkpoint_callback:
        raise ValueError(
            "You're trying to resume training, but checkpointing is not enabled."
        )

    skip = False
    with mlflow.start_run(
        experiment_id=experiment.experiment_id,
        run_name=mlflow_conf.run_name,
        run_id=run_id,
        nested=(run_id is not None),
    ):

        if run_id is None:
            params_to_log = mlflow_conf.get("params_to_log", None)
            if params_to_log is not None:
                mlflow.log_params(
                    flatten_config(OmegaConf.to_container(params_to_log, resolve=True))
                )

        if run_id is not None:
            client = MlflowClient(tracking_uri=mlflow_conf["tracking_uri"])
            run = client.get_run(run_id)
            if patience is not None:
                if "wait_count" in run.data.metrics:
                    if run.data.metrics["wait_count"] >= patience:
                        logger.info(
                            "This model has trained until or beyond the early "
                            "stopping patience value. Skipping"
                        )
                        skip = True

            if trainer.max_epochs is not None:
                if "train_loss" in run.data.metrics:
                    timepoints = client.get_metric_history(run_id, "train_loss")
                    if len(timepoints) >= trainer.max_epochs:
                        logger.info(
                            "This model has trained until or beyond "
                            "max epochs. Skipping"
                        )
                        skip = True

        if not skip:
            # we don't know yet if there's a checkpoint
            ckpt_path = None
            with tempfile.TemporaryDirectory() as tmp_dir:
                if run_id is not None:
                    logger.info("Trying to retrieve checkpoint to resume training")
                    ckpt_path = _get_latest_checkpoint(
                        mlflow_conf.tracking_uri, run_id, tmp_dir
                    )

                _log_conf(tmp_dir, full_conf, "train")

                logger.info("Calling trainer.fit")
                trainer.fit(model, data, ckpt_path=ckpt_path)

                if test:
                    logger.info("Calling trainer.test")
                    trainer.test(model, data)
                if test:
                    logger.info("Calling trainer.predict")
                    trainer.predict(model, data)

        mlflow.end_run(status="FINISHED")


def mlflow_test(mlflow_conf, trainer, data, full_conf):
    experiment, run_id, _ = _mlflow_prep(mlflow_conf, trainer, "test")

    if run_id is None:
        raise ValueError(
            "You're calling serotiny test but you haven't specified the run_id"
        )

    with mlflow.start_run(
        experiment_id=experiment.experiment_id,
        run_name=mlflow_conf.run_name,
        run_id=run_id,
        nested=(run_id is not None),
    ):

        with tempfile.TemporaryDirectory() as tmp_dir:
            if _get_config(
                mlflow_conf.tracking_uri, run_id, tmp_dir
            ) is not None and not full_conf.get("force", False):
                logger.info(
                    "This run has been tested before. If you want to "
                    "retest it, set ++force=True"
                )
            else:
                model = load_model_from_checkpoint(mlflow_conf.tracking_uri, run_id)

                _log_conf(tmp_dir, full_conf, "test")

                if model is not None:
                    logger.info("Calling trainer.test")
                    trainer.test(model, data)
        mlflow.end_run(status="FINISHED")


def mlflow_predict(mlflow_conf, trainer, data, full_conf):
    experiment, run_id, _ = _mlflow_prep(mlflow_conf, trainer, "predict")

    if run_id is None:
        raise ValueError(
            "You're calling serotiny predict but you " "haven't specified the run_id"
        )

    with mlflow.start_run(
        experiment_id=experiment.experiment_id,
        run_name=mlflow_conf.run_name,
        run_id=run_id,
        nested=(run_id is not None),
    ):

        with tempfile.TemporaryDirectory() as tmp_dir:
            model = load_model_from_checkpoint(mlflow_conf.tracking_uri, run_id)

            _log_conf(tmp_dir, full_conf, "predict")

            if model is not None:
                logger.info("Calling trainer.predict")
                preds = trainer.predict(model, data)

                preds_dir = Path(tmp_dir) / "predictions"
                preds_dir.mkdir(exist_ok=True)

                save_model_predictions(model, preds, preds_dir)
                _log_predictions(preds_dir, mlflow_conf["tracking_uri"], run_id)
        mlflow.end_run(status="FINISHED")


@contextmanager
def upload_artifacts(artifact_path, exclude=[], globstr="*"):
    """Util context manager to upload artifacts.

    It yields a temporary directory to store results and then uploads the files
    stored therein. Optionally exclude given files or restrict to a given glob
    filter
    """

    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            yield tmp_dir
        finally:
            for fpath in Path(tmp_dir).glob(globstr):
                if fpath in exclude or str(fpath) in exclude or not fpath.is_file():
                    continue

                mlflow.log_artifact(local_path=fpath, artifact_path=artifact_path)


@contextmanager
def download_artifact(artifact_path):
    """Util context manager to download artifacts.

    Returns a path to the downloaded artifact
    """

    with tempfile.TemporaryDirectory() as tmp_dir:
        try:
            client = mlflow.tracking.MlflowClient(mlflow.get_tracking_uri())
            yield client.download_artifacts(
                run_id=mlflow.active_run().info.run_id,
                path=artifact_path,
                dst_path=tmp_dir,
            )

        finally:
            pass



import os
from typing import Callable, Optional, Sequence, Union, Type

import numpy as np

from serotiny.io.image import image_loader

from .abstract_loader import Loader


class LoadImage(Loader):
    """Loader class, used to retrieve images from paths given in a dataframe
    column."""

    def __init__(
        self,
        column: str,
        file_type: str = "tiff",
        select_channels: Optional[Sequence] = None,
        transform: Optional[Callable] = None,
        reader: Optional[str] = None,
        dtype: Optional[Union[str, Type[np.number]]] = None,
        load_as_torch: bool = True,
        use_cache: bool = False,
    ):
        """
        Parameters
        ----------
        column: str
            Dataframe column which contains the image path

        file_type: str
            File format of the image. For now, "tiff" is the
            only supported format. But in the future other formats (like zarr)
            might be supported too

        select_channels: Optional[Sequence] = None
            List of channels to include in the loaded image.

        transform: Optional[Callable] = None
            Transform to apply upon loading the image.

        reader: Optional[str] = None
            `aicsimageio` reader to use

        dtype: np.dtype = np.float32
            dtype to use. see https://numpy.org/doc/stable/reference/generated/numpy.ndarray.astype.html  # noqa
            for more info on this

        load_as_torch: bool = True
            Whether to load the image as a torch tensor rather than a numpy
            array. Some transforms require this.

        use_cache: bool = True
            Whether to cache images after downloading them once

        """
        super().__init__()
        self.column = column
        self.select_channels = select_channels
        self.reader = reader
        self.load_as_torch = load_as_torch

        if file_type not in ("tiff"):
            raise NotImplementedError(f"File type {file_type} not supported.")

        self.file_type = file_type
        self.dtype = dtype
        self.transform = transform
        self.use_cache = use_cache

    def _get_cached_path(self, path):
        if not self.use_cache:
            return path

        conf_path = os.getenv("FSSPEC_CONFIG_DIR")
        if conf_path is not None:
            if "simplecache::" not in str(path):
                return "simplecache::" + path
        return path

    def __call__(self, row):
        if self.file_type == "tiff":
            return image_loader(
                self._get_cached_path(row[self.column]),
                select_channels=self.select_channels,
                output_dtype=self.dtype,
                transform=self.transform,
                reader=self.reader,
                return_as_torch=self.load_as_torch,
            )



import torch
from torch.distributions import Weibull
from torch.nn.modules.loss import _Loss as Loss


def weibull_log_probs(beta, alpha, target, eps):
    alpha = alpha + eps
    return (
        torch.log(beta)
        - torch.log(alpha)
        + (beta - 1) * (torch.log(target) - torch.log(alpha))
        - beta * (target / alpha)
    )


class WeibullLogLoss(Loss):
    def __init__(self, reduction="sum", mode="explicit", eps=1e-10):
        super(WeibullLogLoss, self).__init__(None, None, reduction)
        self.mode = mode
        self.eps = eps

    def forward(self, a, b, target):

        if self.mode == "explicit":
            log_probs = weibull_log_probs(b, a, target, self.eps)
        else:
            log_probs = Weibull(a, b).log_prob(target)

        # reduce across batch dimension
        if self.reduction == "none":
            return -log_probs
        elif self.reduction == "sum":
            return -log_probs.sum()
        elif self.reduction == "mean":
            return -log_probs.mean()
        else:
            raise NotImplementedError(f"Unavailable reduction type: {self.reduction}")



import logging
from inspect import signature
from pathlib import Path
from typing import Optional, Sequence, T, Union

import pandas as pd
from makefun import wraps

import serotiny.transforms.dataframe as df_transforms
from serotiny.io.dataframe import read_dataframe

from .._utils import PipelineCLI

PathLike = Union[str, Path]
OneOrMany = Union[T, Sequence[T]]


logger = logging.getLogger(__name__)


def _store_one_df(result, output_path):
    if output_path.suffix == ".parquet":
        result.to_parquet(output_path)
    elif output_path.suffix == ".csv":
        result.to_csv(output_path)
    else:
        raise TypeError("output path must be either .parquet or .csv")


def _store_many_dfs(result, output_path):
    for name, df in result.items():
        assert isinstance(name, str)
        assert isinstance(df, pd.DataFrame)

        name = Path(name)
        if name.suffix == ".csv":
            df.to_csv(output_path / name)
        elif name.suffix == ".parquet":
            df.to_parquet(output_path / name)
        elif name.suffix == "":
            output_path.mkdir(parents=True, exist_ok=True)
            df.to_csv(output_path / name.with_suffix(".csv"))
        else:
            raise TypeError(f"Unexpected suffix: '{name.suffix}'")


def _read_one_or_many(
    input_manifests: OneOrMany[PathLike], merge_col: Optional[str] = None
):
    input_manifest = None

    if isinstance(input_manifests, (str, Path)):
        input_manifests = [input_manifests]

    for manifest in input_manifests:
        manifest = Path(manifest)
        if not manifest.exists():
            raise FileNotFoundError(f"Given input path {manifest} does not exist")
        manifest = read_dataframe(manifest)

        if input_manifest is None:
            input_manifest = manifest
            continue

        if merge_col is not None:
            input_manifest = input_manifest.merge(manifest, on=merge_col, how="outer")
        else:
            input_manifest = input_manifest.concat(manifest)

    return input_manifest


def _dataframe_from_disk(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        func_sig = signature(func)
        func_arg_names = list(func_sig.parameters.keys())

        base_args = {}
        for arg_ix, arg_value in enumerate(args):
            base_args[func_arg_names[arg_ix]] = arg_value

        base_args.update(kwargs)

        if "dataframe" in base_args:
            if not isinstance(base_args["dataframe"], pd.DataFrame):
                if base_args["dataframe"] not in (..., "..."):
                    base_args["dataframe"] = read_dataframe(base_args["dataframe"])

        return func(**base_args)

    return wrapper


class DataframeTransformCLI(PipelineCLI):
    """Apply a transform (or chain of transforms) to a dataframe."""

    @classmethod
    def _decorate(cls, func):
        return _dataframe_from_disk(func)

    def __init__(self, input_manifests=None, output_path=None):
        super().__init__(
            output_path=output_path,
            transforms=[
                df_transforms.split_dataframe,
                df_transforms.filter_rows,
                df_transforms.filter_columns,
                df_transforms.sample_n_each,
                df_transforms.append_one_hot,
                df_transforms.append_labels_to_integers,
                df_transforms.append_class_weights,
                df_transforms.make_random_df,
            ],
            store_methods={
                pd.DataFrame: _store_one_df,
                dict: _store_many_dfs,
            },
        )

        self._result = (
            None if input_manifests is None else _read_one_or_many(input_manifests)
        )



from torch.utils.data import Dataset


class _Row:
    """Helper class to enable string indexing of numpy arrays."""

    def __init__(self, array, columns):
        self.array = array
        self.columns = columns
        self.index = self.columns

    def __getitem__(self, col):
        if isinstance(col, (list, tuple)):
            return self.array[[self.columns.index(_col) for _col in col]]
        return self.array[self.columns.index(col)]

    def __getattr__(self, col):
        return self.array[self.columns.index(col)]


class DataframeDataset(Dataset):
    """Class to wrap a pandas DataFrame in a pytorch Dataset. In practice, at AICS we
    use this to wrap manifest dataframes that point to the image files that correspond
    to a cell. The `loaders` dict contains a loading function for each key, normally
    consisting of a function to load the contents of a file from a path.

    Parameters
    ----------
    dataframe: pd.DataFrame
        The file which points to or contains the data to be loaded

    loaders: dict
        A dict which contains methods to appropriately load data from columns
        in the dataset.
    """

    def __init__(self, dataframe, loaders):

        # store only the numpy array containing all the values. this is because
        # in multiprocessing settings, having a pandas dataframe here can incur
        # in excessive memory use, because of copy-on-write memory sharing mechanisms,
        # triggered by pandas' inner workings
        self.dataframe = dataframe.values
        self.columns = dataframe.columns.tolist()

        self.loaders = loaders

    def __len__(self):
        return len(self.dataframe)

    def __getitem__(self, idx):
        row = _Row(self.dataframe[idx], self.columns)

        return {key: loader(row) for key, loader in self.loaders.items()}



import math

import torch
import torch.nn as nn


def spatial_pyramid_pool(x, out_pool_sizes):

    input_dims = x.shape[2:]
    batch_size = x.shape[0]

    for i in range(len(out_pool_sizes)):
        wids = []
        pads = []

        for dim in input_dims:
            wid = int(math.ceil(dim / out_pool_sizes[i]))
            wids.append(wid)
            pad = int(math.floor((wid * out_pool_sizes[i] - dim + 1) / 2))
            pads.append(pad)

        if len(input_dims) == 1:
            maxpool = nn.MaxPool1d
        elif len(input_dims) == 2:
            maxpool = nn.MaxPool2d
        elif len(input_dims) == 3:
            maxpool = nn.MaxPool3d
        else:
            raise ValueError("Spatial Pyramid Pool only supports 1, 2, or 3d data")
        maxpool = maxpool(tuple(wids), stride=tuple(wids), padding=tuple(pads))

        out = maxpool(x)
        if i == 0:
            spp = out.view(batch_size, -1)
        else:
            spp = torch.cat((spp, out.view(batch_size, -1)), 1)

    return spp


class SpatialPyramidPool(nn.Module):
    def __init__(self, out_pool_sizes):
        super().__init__()
        self.out_pool_sizes = out_pool_sizes

    def forward(self, x):
        return spatial_pyramid_pool(x, self.out_pool_sizes)



from typing import Sequence, Union, Optional, Callable

import numpy as np
import torch
import torch.nn as nn
from torch.nn.modules.loss import _Loss as Loss

from .base_model import BaseModel

Array = Union[torch.Tensor, np.ndarray, Sequence[float]]


class BasicModel(BaseModel):
    """A minimal Pytorch Lightning wrapper around generic Pytorch models."""

    def __init__(
        self,
        network: nn.Module,
        loss: Loss,
        x_label: str = "x",
        y_label: str = "y",
        optimizer: torch.optim.Optimizer = torch.optim.Adam,
        save_predictions: Optional[Callable] = None,
        fields_to_log: Optional[Sequence] = None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        network: nn.Module
            The network to wrap
        loss: Loss
            The loss function to optimize for
        x_label: str = "x"
            The key used to retrieve the input from dataloader batches
        y_label: str = "y"
            The key used to retrieve the target from dataloader batches
        optimizer: torch.optim.Optimizer = torch.optim.Adam
            The optimizer class
        save_predictions: Optional[Callable] = None
            A function to save the results of `serotiny predict`
        fields_to_log: Optional[Union[Sequence, Dict]] = None
            List of batch fields to store with the outputs. Use a list to log
            the same fields for every training stage (train, val, test, prediction).
            If a list is used, it is assumed to be for test and prediction only
        """
        super().__init__()
        self.network = network
        self.loss = loss

        self._squeeze_y = False

        if save_predictions is not None:
            self.save_predictions = save_predictions
        self.fields_to_log = fields_to_log

    def parse_batch(self, batch):
        return (batch[self.hparams.x_label], batch[self.hparams.y_label])

    def forward(self, x, **kwargs):
        return self.network(x, **kwargs)

    def _step(self, stage, batch, batch_idx, logger):
        x, y = self.parse_batch(batch)

        yhat = self.forward(x)

        if self._squeeze_y:
            loss = self.loss(yhat, y.squeeze())
        else:
            try:
                loss = self.loss(yhat, y)
            except RuntimeError as err:
                if y.shape[-1] == 1:
                    try:
                        loss = self.loss(yhat, y.squeeze())
                        self._squeeze_y = True
                    except Exception as inner_err:
                        raise inner_err from err
                else:
                    raise err

        if stage != "predict":
            self.log(
                f"{stage}_loss",
                loss.detach(),
                logger=logger,
                on_step=True,
                on_epoch=True,
            )

        output = {
            "loss": loss,
            "yhat": yhat.detach().squeeze(),
            "y": y.detach().squeeze(),
        }

        if isinstance(self.fields_to_log, list):
            if stage in ["predict", "test"]:
                for field in self.fields_to_log:
                    output[field] = batch[field]

        elif isinstance(self.fields_to_log, dict):
            if stage in self.fields_to_log:
                for field in self.fields_to_log[stage]:
                    output[field] = batch[field]

        return output



import numpy as np
import torch


class NormalizeAbsolute:
    def __init__(self, axis=-1, order=2):
        self.axis = axis
        self.order = order

    def __call__(self, a):
        l2 = np.atleast_1d(np.linalg.norm(a, self.order, self.axis))
        l2[l2 == 0] = 1
        return a / np.expand_dims(l2, self.axis)


class NormalizeMean:
    def __init__(self, scale=1.0):
        self.scale = scale

    def __call__(self, img, per_dim=None):
        """Subtract mean, set STD to 1.0.

        Parameters:
          per_dim: normalize along other axes dimensions not equal to per dim
        """
        axis = tuple([i for i in range(img.ndim) if i != per_dim])
        slices = tuple(
            [slice(None) if i == per_dim else np.newaxis for i in range(img.ndim)]
        )  # to handle broadcasting
        result = img.astype(np.float32)
        result -= np.mean(result, axis=axis)[slices]
        result /= np.std(result, axis=axis)[slices]
        result *= self.scale
        return result


class NormalizeMinMax:
    def __init__(
        self, clip_min=None, clip_max=None, clip_quantile=False, return_torch=False
    ):
        self.clip_min = clip_min
        self.clip_max = clip_max
        self.clip_quantile = clip_quantile
        self.return_torch = return_torch

    def __call__(self, img):
        if self.clip_min is not None:
            if isinstance(self.clip_min, (int, float)):
                if not self.clip_quantile:
                    clip_min = [self.clip_min] * img.shape[0]
                else:
                    clip_min = [
                        np.quantile(img[ch], self.clip_min)
                        for ch in range(img.shape[0])
                    ]
            else:
                clip_min = self.clip_min
        else:
            clip_min = [None] * img.shape[0]
        assert len(clip_min) == img.shape[0]

        if self.clip_max is not None:
            if isinstance(self.clip_max, (int, float)):
                if not self.clip_quantile:
                    clip_max = [self.clip_max] * img.shape[0]
                else:
                    clip_max = [
                        np.quantile(img[ch], self.clip_max)
                        for ch in range(img.shape[0])
                    ]
            else:
                clip_max = self.clip_max
        else:
            clip_max = [None] * img.shape[0]
        assert len(clip_max) == img.shape[0]

        for chan in range(img.shape[0]):
            if clip_min[chan] is not None:
                _chan_result = np.where(
                    img[chan] < clip_min[chan],
                    clip_min[chan],
                    img[chan],
                )
                if isinstance(img, torch.Tensor):
                    _chan_result = torch.tensor(_chan_result)
                img[chan] = _chan_result

            if clip_max[chan] is not None:
                _chan_result = np.where(
                    img[chan] > clip_max[chan],
                    clip_max[chan],
                    img[chan],
                )
                if isinstance(img, torch.Tensor):
                    _chan_result = torch.tensor(_chan_result)
                img[chan] = _chan_result

            m = img[chan].min()
            M = img[chan].max()
            img[chan] = (img[chan] - m) / (M - m)

        if self.return_torch:
            img = torch.tensor(img)
        return img
